! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! Module: binadaptmerge_mod
!
! Authors:  Brendan Perry - NASA/GSFC (brendan.h.perry@nasa.gov)
!            Comprehensive rewrite, merger and modularization for SAS-21
!           Steven L. Snowden - NASA/GSFC (original algorithm)
!           Kip Kuntz - NASA/GSFC/JHU (original algorithm)
!
! Description: Extended Source Analysis Software
!
! Comments:
!  Former ESAS tasks adaptmerge and binimagemerge were merged, just as
!  former ESAS tasks adapt and binimage were merged into binadapt. 
!
! $Id: binadaptmerge_mod.f90,v 1.2 2009/02/20 13:27:08 aibarra Exp $
!-
!
!  SAS Version 13.0
!  Version 1.0  2 November 2012
!

! BHP: Part of the Extended Source Analysis Software collection of SAS tasks.
! This task is now a merged binimagemerge and adaptmerge from pre-SAS-21
! ESAS. The two tasks shared 90% of the same functionality, so they were merged
! and the differences are handled by the parameters.  Creates adaptively smoothed
! and/or binned, exposure corrected and background subtracted images for the
! XMM-ESAS package.
!
module binadaptmerge_mod

  use dal
  use param 
  use errorHandling
  use esas_mod      ! in goflib, formerly in subsn (SS)
  use gofutils_mod  ! in goflib, formerly in marq_mod
  use gofcoords_mod ! in goflib, formerly in subsn (SS)
  use gofio_mod     ! in goflib, formerly in subsn (SS)

  implicit none 

  private

  character(len=*), parameter :: &
       elowpar      = "elowlist", &       ! list of elo bounds
       ehighpar     = "ehighlist", &      ! list of ehi bounds 
       partpar      = "withpartbkg", &    ! QPB particle background
       sppar        = "withspbkg", &      ! soft proton background
       swcxpar      = "withswcxbkg", &    ! swcx background
       maskpar      = "withmask", &       ! perform masking?
       maskfilepar  = "maskfile",&        ! mask filename 
       withbinpar   = "withbinning", &    ! perform binning? 
       binfactpar   = "binfactor", &      ! binfactor
       withsmoopar  = "withsmoothing", &  ! perform smoothing
       smoocntspar  = "smoothcounts", &   ! smoothingcounts
       tmpar        = "maskthresh", &     ! threshold for masking 
       fillpar      = "fill"!, &           ! Number of passes to fill in empty pixels
  !xdimpar      = "xdim", &           ! input file X dimension
  !ydimpar      = "ydim"              ! input file Y dimension


  public binadaptmerge

contains

  !---------------------------------------------------------------------
  subroutine binadaptmerge
    !---------------------------------------------------------------------

    type(DataSetT) :: in_S, outADAPT_S, outRATE_S, outSIZE_S, outSIGMA_S, &
         inMASK_S, inQPB_S, inPROT_S, inSWCX_S, pnOOT_S, inEXP_S, inBKG_S
    type(ArrayT)   :: in_A, outADAPT_A, outRATE_A, outSIZE_A, outSIGMA_A, &
         inMASK_A, inQPB_A, inPROT_A, inSWCX_A, pnOOT_A, inEXP_A, inBKG_A
    type(TableT)   :: pnOOT_T, inPROT_T, inSWCX_T, inBKG_T, &
         inEXP_T, in_T, inMASK_T
    character (len=256) :: in_F, outADAPT_F, outRATE_F, outSIZE_F, outSIGMA_F, &
         inMASK_F, inQPB_F, inPROT_F, inSWCX_F, pnOOT_F, inEXP_F, inBKG_F, &
         outSIZEQ_F, outRADQ_F
    character (len=25) :: smode
    character (len=10) :: obsid,  cunitO(2), ctypeO(2), radecsysO, cunit(2), &
         ctype(2), cunitE(2), ctypeE(2), radecsysE, radecsys, style, units
    character (len=7)  :: name
    character (len=6)  :: filter, prefix
    character (len=5), allocatable, dimension(:) :: celow, cehigh
    character (len=5)  :: expid, exptype, suffix
    character (len=2)  :: inst
    !
    integer :: xSize, ySize, xSizeD, ySizeD, xSizeI, ySizeI, xdim, ydim, xdimo, ydimo
    integer :: i, j, k, ii, jj, i1, i2, j1, j2, jj1, jj2, ib, peel
    integer :: status, ios, ibin, ir, irr, ioff, outlun, dims(2)
    integer :: type, frmt, revl, nR, el, eh
    integer :: binfact, typeA, fill, nBandsL, nBandsH, nBands
    integer, allocatable, dimension (:) :: n, elow, ehigh
    integer, dimension(:), pointer :: dimVector
    integer(kind=int16), dimension(:,:), pointer :: aDI16
    integer(kind=int32), dimension(:,:), pointer :: aDI32, inFOV_JP, pnOOT_JP, &
         inMASK_JP
    !
    logical :: doit, oot, warndie, withpart, withprot, withswcx, withmask, withbin, &
         withsmoo, withoot, withcomb
    !
    real :: backO, areaO, expoO, backD, areaD, expoD, smoofact
    real :: imcounts, qpbcounts, expcounts, protcounts, swcxcounts
    real :: maskthresh, area, bscale, bzero
    real :: xcen, ycen ! why reals???!!!
    real :: b, blast, c, clast, e, elast, ex, num, numlast, ootsca, r, &
         rad, s, sca, temp, t, tn, tlim, trad, xmax, ymax
    real(kind=double) :: crvalO(2), cdeltO(2), crpixO(2), crotaO(2), &
         ra_pntO, dec_pntO, pa_pntO, crval(2), cdelt(2), crpix(2), crota(2), &
         ra_pnt, dec_pnt, pa_pnt, crvalE(2), cdeltE(2), crpixE(2), crotaE(2), &
         ra_pntE, dec_pntE, pa_pntE
    real, dimension(:), allocatable :: br, cr, er, rn, sr, xr, yr 
    real, dimension(:), allocatable :: bck, cts, expos, x, y, y2, ys ! why 50???!!!
    real, dimension(:), allocatable :: xh, yh ! why 10000/1000 in orig code???!!!
    real, dimension(:,:), allocatable :: mask, expose, back, counts, ootcounts, &
         prot, swcx, pnOOT, adapted, sized, rate, sigma
    real(kind=single), dimension(:,:), pointer :: aDR32, inFOV_RP, inMASK_RP, &
         inQPB_RP, inPROT_RP, inSWCX_RP, inEXP_RP, outADAPT_RP, outRATE_RP, outSIZE_RP, &
         outSIGMA_RP, inBKG_RP, inOOT_RP, in_RP
    real(kind=double), dimension(:,:), pointer :: aDR64
    !
    allocate(xh(1000),yh(1000)) ! why 1000???!!!
    xh = 0.
    yh = 0.
    do i=1,1000
       xh(i) = -0.1 + i/15.0
       yh(i) = 0.0
    enddo
    !
    ! SS: Set the scale factor for the maps, this is the ratio of
    !     one square degree to the solid angle of the original map pixel
    !
    sca = (60.0*60.0)*(60.0*60.0)/(2.5*2.5)
    !
    nBandsL  = parameterCount(elowpar)
    nBandsH  = parameterCount(ehighpar)
    if (nBandsL .ne. nBandsH) then
       write(errstr,*)"Number of elow and ehigh values must be equal, exiting..."
       call error("eLowsNEeHighs",trim(errstr))
    endif
    nBands = nBandsL ! fix this???!!!
    !
    ! unlike binadapt, can have many bands, not just 2 ???!!!
    allocate(elow(nbands),ehigh(nbands),celow(nbands),cehigh(nbands))
    elow   = 0
    ehigh  = 0
    celow  = ""
    cehigh = ""
    do i=1,nBands
       elow(i)  = intParameter(elowpar,i-1)
       ehigh(i) = intParameter(ehighpar,i-1)
       write(celow(i),'(I5)') elow(i)
       write(cehigh(i),'(I5)') ehigh(i)
    end do
    ! takes care of singular band
    if (nBands .eq. 1) then 
       celow(2)  = celow(1)
       cehigh(2) = cehigh(1)
       elow(2)   = elow(1)
       ehigh(2)  = ehigh(1)
    end if
    !
    withpart    = booleanParameter(partpar)     ! OLDNEW: IPART
    withprot    = booleanParameter(sppar)       ! OLDNEW: ISOFT
    withswcx    = booleanParameter(swcxpar)     ! OLDNEW: ISWCX
    !
    withmask    = .FALSE.                       ! is by default anyway
    withmask    = booleanParameter(maskpar)     ! OLDNEW: IMASK
    inMASK_F    = "" 
    if (withmask) inMASK_F   = stringParameter(maskfilepar) ! OLDNEW: infile! grrr....
    !
    withbin     = .FALSE.
    withbin     = booleanParameter(withbinpar)
    if (withbin) then
       binfact = intParameter(binfactpar)   ! OLDNEW: ibin
    else
       binfact = 1
    endif
    !
    withsmoo    = .FALSE. ! new because merging binmerge and adaptmerge like binimage and adaptimage
    withsmoo    = booleanParameter(withsmoopar)
    if (withsmoo) then
       smoofact = realParameter(smoocntspar)  ! OLDNEW: SN
    else
       smoofact = 1.
    endif
    !
    if ((.not. withbin) .and. (.not. withsmoo)) then
       call warning("noBinNoSmooth",&
            "withbinning=F and withsmoothing=F is an odd choice unless simply combining inputs")
    endif
    if((binfact .eq. 1) .and. (smoofact .eq. 1)) then
       call warning("noBinNoSmooth",&
            "binfactor=1 and smoothfactor=1 is an odd choice unless simply combining inputs")
    endif
    !
    maskthresh = realParameter(tmpar) ! OLDNEW: temsc
    fill = intParameter(fillpar)      ! OLDNEW: IFILL/fill
    ! BHP: maybe later....
    !    xdim = intParameter(xdimpar)      ! from first file before
    !    ydim = intParameter(ydimpar)      ! from first file before
    !    getdims = .FALSE.
    ! BHP: create output FITS files
    ! BHP: maybe tweak filenames again???!!!
    ! CREATE OUTPUT FILES:
    ! OLDNEW: adapt-350-1100.fits/merged-adapt-350-1100.fits
    ! OLDNEW: size-350-1100.fits/merged-size-350-1100.fits
    ! BHP: create output fits filenames 
    ! BHP: or just do all of them by default???!!!
    prefix = "merged"
    style  = "adapt"
    suffix = ".fits"
    call makeFileNames(prefix,suffix,style,celow(1),cehigh(nBands),outADAPT_F)
    prefix = "merged"
    style  = "size"
    suffix = ".fits"
    call makeFileNames(prefix,suffix,style,celow(1),cehigh(nBands),outSIZE_F)
    prefix = "merged"
    style  = "rate"
    suffix = ".fits"
    call makeFileNames(prefix,suffix,style,celow(1),cehigh(nBands),outRATE_F)
    prefix = "merged"
    style  = "sigma"
    suffix = ".fits"
    call makeFileNames(prefix,suffix,style,celow(1),cehigh(nBands),outSIGMA_F)
    !
    ! BHP: and now create the QDP plot names
    ! OLDNEW: adapt-350-1100.qdp/merged-adapt-350-1100.qdp
    ! OLDNEW: size-350-1100.qdp/merged-size-350-1100.qdp
    !
    prefix = "merged"
    style  = "radfilt"
    suffix = ".qdp"
    call makeFileNames(prefix,suffix,style,celow(1),cehigh(nBands),outRADQ_F)
    prefix = "merged"
    style  = "size"
    suffix = ".qdp"
    call makeFileNames(prefix,suffix,style,celow(1),cehigh(nBands),outSIZEQ_F)
    !
    ! SS: Find the size of the input arrays
    ! create the filenames, fill the arrays
    ! BHP: call check_clobber ???!!!
    !
    xdimo = 0
    ydimo = 0
    do i=1,nBandsL
       ! FOVIM
       prefix = "mosaic" ! from mosaicmerge (only???!!!)
       style  = "fovim"  ! OLDNEW: obj-im-350-1100.fits/mosaic-fovim-350-1100.fits
       suffix = ".fits"
       call makeFileNames(prefix,suffix,style,celow(i),cehigh(i),in_F)
       write(errstr,*)'***** Reading file: ',trim(adjustl(in_F))
       call message(AppMsg,Always)
       ! BHP: all input arrays below are R32
       ! BHP: extract FOV image array
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_open_set(in_F,in_S,1,status,warndie) ! goflib/gofio_mod
       type = 3 ! image
       call read_2Darray(in_S,in_A,in_RP,aDR64,aDI16,aDI32,xdim,ydim,typeA)
       ! put in sanity check for xsize/ysize???!!!
       if (i .eq. 1) then
          allocate(counts(xdim,ydim))
          allocate(back(xdim,ydim))
          allocate(expose(xdim,ydim))
          allocate(mask(xdim,ydim))
          allocate(prot(xdim,ydim))
          allocate(swcx(xdim,ydim))
          counts = 0.
          back   = 0.
          expose = 0.
          mask   = 0.
          prot   = 0.
          swcx   = 0.
          ! allocate(off(xdim,ydim)) where's OUTLAY???!!!
          xdimo = xdim / binfact
          ydimo = ydim / binfact
          ioff  = mod(xSizeI,binfact)/2 
          write(errstr,*) "Input Dimensions (from 1st image): ", xdim, ydim
          call message(AppMsg,Always)
          write(errstr,*) "Output Dimensions                : ", xdimo, ydimo
          call message(AppMsg,Always)
       endif
       do j=1,xdim
          do k=1,ydim
             counts(j,k) = counts(j,k) + in_RP(j-1,k-1)
             !write(33,*)j,k,counts(j,k)
          enddo
       enddo
       nullify (in_RP)    ! is this necessary if doing call release???!!!
       call release(in_S) ! free mosaic-fovim-elow-ehigh.fits
       !
       ! SS:  Enter the exposure array file names and read in the data
       !
       ! EXPIM
       prefix = "mosaic" ! from mosaicmerge (only???!!!)
       style  = "expim"  ! OLDNEW: exp-im-350-1100.fits/mosaic-expim-350-1100.fits
       suffix = ".fits"
       call makeFileNames(prefix,suffix,style,celow(i),cehigh(i),in_F)
       write(errstr,*)'***** Reading file: ',trim(adjustl(in_F))
       call message(AppMsg,Always)
       ! BHP: all input arrays below are R32
       ! BHP: extract EXPOS image array
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_open_set(in_F,in_S,1,status,warndie) ! goflib/gofio_mod
       type = 3 ! image
       call read_2Darray(in_S,in_A,in_RP,aDR64,aDI16,aDI32,xSizeI,ySizeI,typeA)
       warndie = .FALSE.
       call coord_keys (in_S,ra_pntE,dec_pntE,pa_pntE,crvalE,cdeltE, &
            cunitE,ctypeE,crpixE,crotaE,radecsysE,warndie)
       ! put in sanity check for xsize/ysize???!!!
       do j=1,xdim
          do k=1,ydim
             expose(j,k) = expose(j,k) + in_RP(j-1,k-1)
          enddo
       enddo
       nullify (in_RP)    ! is this necessary if doing call release???!!!
       call release(in_S) ! free mosaic-expim-elow-ehigh.fits
       if (nbands .gt. 1) then 
          do j=1,xdim
             do k=1,ydim
                expose(j,k) = expose(j,k)/nbands
             enddo
          enddo
       endif
       ! BHP: now do the rest of the input types as necessary
       !
       ! SS: Enter the background count array file names and read in the data
       !
       ! BKGIM
       if(withpart) then 
          prefix = "mosaic" ! from mosaicmerge (only???!!!)
          style  = "bkgim"  ! OLDNEW: back-im-350-1100.fits/mosaic-bkgim-350-1100.fits
          suffix = ".fits"
          call makeFileNames(prefix,suffix,style,celow(i),cehigh(i),in_F)
          write(errstr,*)'***** Reading file: ',trim(adjustl(in_F))
          call message(AppMsg,Always)
          ! BHP: all input arrays below are R32
          ! BHP: extract EXPOS image array
          status = 0
          warndie = .TRUE. ! croak if not found or unreadable
          call check_open_set(in_F,in_S,1,status,warndie) ! goflib/gofio_mod
          type = 3 ! image
          call read_2Darray(in_S,in_A,in_RP,aDR64,aDI16,aDI32,xSizeI,ySizeI,typeA)
          ! put in sanity check for xsize/ysize???!!!
          do j=1,xdim
             do k=1,ydim
                back(j,k) = back(j,k) + in_RP(j-1,k-1)
             enddo
          enddo
          nullify (in_RP)    ! is this necessary if doing call release???!!!
          call release(in_S) ! free mosaic-expim-elow-ehigh.fits
       endif
       !
       ! SS: Enter the soft proton count array file names and read in the data
       !
       ! PROTIM
       if(withprot) then 
          prefix = "mosaic" ! from mosaicmerge (only???!!!)
          style  = "protim"  ! OLDNEW: prot-im-350-1100.fits/mosaic-protim-350-1100.fits
          suffix = ".fits"
          call makeFileNames(prefix,suffix,style,celow(i),cehigh(i),in_F)
          write(errstr,*)'***** Reading file: ',trim(adjustl(in_F))
          call message(AppMsg,Always)
          ! BHP: all input arrays below are R32
          ! BHP: extract EXPOS image array
          status = 0
          warndie = .TRUE. ! croak if not found or unreadable
          call check_open_set(in_F,in_S,1,status,warndie) ! goflib/gofio_mod
          type = 3 ! image
          call read_2Darray(in_S,in_A,in_RP,aDR64,aDI16,aDI32,xSizeI,ySizeI,typeA)
          ! put in sanity check for xsize/ysize???!!!
          do j=1,xdim
             do k=1,ydim
                prot(j,k) = prot(j,k) + in_RP(j-1,k-1)
             enddo
          enddo
          nullify (in_RP)    ! is this necessary if doing call release???!!!
          call release(in_S) ! free mosaic-expim-elow-ehigh.fits
       endif
       !
       ! SS: Enter the soft proton offset backgrpount count array file names and 
       !     read in the data
       !
       ! SWCXIM
       if(withswcx) then 
          prefix = "mosaic" ! from mosaicmerge (only???!!!)
          style  = "swcxim"  ! OLDNEW: swcx-im-350-1100.fits/mosaic-swcxim-350-1100.fits
          suffix = ".fits"
          call makeFileNames(prefix,suffix,style,celow(i),cehigh(i),in_F)
          write(errstr,*)'***** Reading file: ',trim(adjustl(in_F))
          call message(AppMsg,Always)
          ! BHP: all input arrays below are R32
          ! BHP: extract EXPOS image array
          status = 0
          warndie = .TRUE. ! croak if not found or unreadable
          call check_open_set(in_F,in_S,1,status,warndie) ! goflib/gofio_mod
          type = 3 ! image
          call read_2Darray(in_S,in_A,in_RP,aDR64,aDI16,aDI32,xSizeI,ySizeI,typeA)
          ! put in sanity check for xsize/ysize???!!!
          do j=1,xdim
             do k=1,ydim
                swcx(j,k) = swcx(j,k) + in_RP(j-1,k-1)
             enddo
          enddo
          nullify (in_RP)    ! is this necessary if doing call release???!!!
          call release(in_S) ! free mosaic-expim-elow-ehigh.fits
       endif
       ! SS: Enter the soft proton offset backgrpount count array file names and 
       !     read in the data
       ! These are the 'offset-im-elow-ehigh.fits' files from merge_off_xmm produced
       ! by OUTLAY, which is nowhere to be found (summer, 2022)

    end do ! end loop through all input files
    !
    !  SS: Add the background components
    !
    do i=1,xdim
       do ii=1,ydim
          back(i,ii) =  back(i,ii) + prot(i,ii) + swcx(i,ii) ! + off(i,ii) ???!!!
       enddo
    enddo
    
    ! SS:  Read in the template
    ! MASKIM
    if(withmask) then 
       write(errstr,*)'***** Reading file: ',trim(adjustl(inMASK_F))
       call message(AppMsg,Always)
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_open_set(in_F,in_S,1,status,warndie) ! goflib/gofio_mod
       type = 3 ! image
       ! SS: Return the mask
       call read_2Darray(in_S,in_A,in_RP,aDR64,aDI16,aDI32,xSizeI,ySizeI,typeA)
       ! put in sanity check for xsize/ysize???!!!
       do j=1,xdim
          do k=1,ydim
             mask(j,k) = mask(j,k) + in_RP(j-1,k-1)
             if (expose(j,k) .le. 0.) mask(j,k) = 0.
          enddo
       enddo
       nullify (in_RP)    ! is this necessary if doing call release???!!!
       call release(in_S) ! free mosaic-expim-elow-ehigh.fits
    else
       do j=1,xdim
          do k=1,ydim
             mask(j,k) = expose(j,k)
             !write(67,*)j,k,mask(j,k),expose(j,k) values are correct
          enddo
       enddo
    endif
    ! BHP: brought over from binadapt to keep track
    imcounts   = 0.0 ! RNC
    qpbcounts  = 0.0 ! RNB
    protcounts = 0.0 ! RNP
    swcxcounts = 0.0 ! RNS
    expcounts  = 0.0 ! REX
    ex = 0.0
    do i=1,xdim
       do ii=1,ydim
          imcounts   = imcounts   + counts(i,ii)
          qpbcounts  = qpbcounts  + back(i,ii)
          protcounts = protcounts + prot(i,ii)
          swcxcounts = swcxcounts + swcx(i,ii)
          expcounts  = expcounts  + expose(i,ii)
          back(i,ii) = back(i,ii) + prot(i,ii) + swcx(i,ii)
          if(expose(i,ii) .gt. ex) then
             ex = expose(i,ii)
             xcen = i
             ycen = ii
          endif
       enddo
    enddo
    ! BHP: these are exactly the same, old v new
    write(errstr,*) "Total Counts   : ", imcounts 
    call message(AppMsg,Always)
    write(errstr,*) "Exposure       : ", expcounts
    call message(AppMsg,Always)
    write(errstr,*) "QPB Counts     : ", qpbcounts
    if (withpart) call message(AppMsg,Always)
    write(errstr,*) "Proton Counts  : ", protcounts
    if (withprot) call message(AppMsg,Always)
    write(errstr,*) "SWCX Counts    : ", swcxcounts
    if (withswcx) call message(AppMsg,Always)
    !
    ! SS: Set the template limit
    !
    j = 0
    tlim = 0.
    do i=1,xdim
       do ii=1,ydim
          if(mask(i,ii) .gt. 0) then
             tlim = tlim + mask(i,ii)
             j = j + 1
          endif
       enddo
    enddo
    
    tlim = maskthresh * tlim / j ! tlim seems correct OLDNEW: temsc/maskthresh

    ! SS: Read in the filling control
    ! NN = NN + 1 BHP: this is set then never used???!!!
    !
    ! SS: Make a radial profile
    !
    allocate(br(1000),cr(1000),er(1000),rn(1000),sr(1000),xr(1000),yr(1000))
    br = 0.
    cr = 0.
    er = 0.
    rn = 0.
    sr = 0.
    xr = 0.
    yr = 0.
    !
    do i=1,xdim
       do ii=1,ydim
          if(mask(i,ii) .gt. tlim) then
             r = sqrt((i-1000.5)**2 + (ii-1000.5)**2) ! 1/2 xdim + 0.5???!!!
             ir = 60.0*r*cdeltE(2) + 1.0
             if(ir .gt. 1000) ir = 1000
             br(ir) = br(ir) + back(i,ii)
             cr(ir) = cr(ir) + counts(i,ii)
             er(ir) = er(ir) + expose(i,ii)
             rn(ir) = rn(ir) + 1.0
            ! write(40,*)i,ii,mask(i,ii),r,ir
            ! write(41,*)i,ii,ir,br(ir),cr(ir),er(ir),rn(ir)
          endif
       enddo
    enddo
    xmax = 0.0
    ymax = 0.0
    do i=1,1000 ! why 1000? xdim/2???!!!
       if(er(i) .gt. 0.0) then
          yr(i) = (cr(i) - br(i))/er(i)
          sr(i) = sqrt(cr(i))/er(i)
          xr(i) = (i - 0.5)
          if((yr(i)) .gt. ymax) ymax = yr(i)
          if(xr(i) .gt. xmax) xmax = xr(i)
       else
          yr(i) = 0.0
          sr(i) = 0.0
          xr(i) = (i - 0.5)/2.0
       endif
    enddo
    !
    ! SS: Write out the new FITS adaptively smoothed image file
    ! SS: Get an unused lun
    !
    ! create and write output QDP file (RADIAL)
    ! e.g. merged-radfilt-350-1100.qdp 
    ! BHP: do the RADIAL plot
    !
    warndie = .FALSE.
    call check_clobber(outRADQ_F,warndie)
    outlun = GetLun()
    open(unit=outlun,status='new',form='formatted',&
         file=trim(adjustl(outRADQ_F)),iostat=ios)
    if (ios .ne. 0) then
       call warning("badQDP",&
            "QDP output file could not be opened for write, no RADIAL QDP output done")
    else
       ymax = 1.05*ymax
       write(outlun,*) 'FONT ROMAN'
       write(outlun,*) 'CS 1.5'
       write(outlun,*) 'READ SERR 2'
       write(outlun,*) 'LA Y Count Rate (counts/s)'
       write(outlun,*) 'LA X Radius'
       write(outlun,*) 'LINE OFF 2'
       write(outlun,*) 'R X 0 ', XMAX
       write(outlun,*) 'R Y 0 ', YMAX
       write(outlun,*) 'LW 2'
       write(outlun,*) 'LA F'
       write(outlun,*) 'LA T Radial Profile'
       do i=1,1000 ! or xdim/binfact???!!!
          write(outlun,*) xr(i), yr(i), sr(i)
       enddo
       close(outlun)
    endif
    i2 = xSizeI / binfact ! OLDNEW: 900/ibin
    allocate(rate(i2,i2),sigma(i2,i2))
    rate  = 0.
    sigma = 0.
    !
    write(errstr,*) "Start the binning"
    call message(AppMsg,Always)
    b = 0.
    c = 0.
    e = 0.
    ! from here ----
    do i=1,i2
       j2 = i*binfact + ioff
       j1 = j2 - binfact + 1
       do ii=1,i2
          jj2 = ii*binfact + ioff
          jj1 = jj2 - binfact + 1
          b = 0.
          c = 0.
          e = 0.
          do j=j1,j2
             do jj=jj1,jj2
                if(mask(j,jj) .gt. tlim) then
                   b = b + back(j,jj)
                   c = c + counts(j,jj)
                   e = e + expose(j,jj)
                endif
             enddo
          enddo
          ! to here ----exactly the same in old adapt and binimage
          if (withsmoo) then ! not done in binimage
             back(i,ii)   = b
             counts(i,ii) = c
             expose(i,ii) = e
          endif
          ! BHP: none of the below was ever done in adapt
          if(e .gt. 0.0) then
             r = sca*(c - b)/e
             if(c .gt. 0.0) then
                s = sca*sqrt(c)/e
             else
                s = 0.0
             endif
          else
             r = 0.0
             s = 0.0
          endif
          rate(i,ii) = r
          sigma(i,ii) = s
       enddo
    enddo
    !
    ! SS: Now smooth the data
    !
    if (withsmoo) then
       allocate(adapted(xdimo,ydimo),sized(xdimo,ydimo)) ! OLDNEW: output/size
       adapted = 0.
       sized   = 0.
       do i=1,xdimo
          if(mod(i,25) .eq. 0) then
             write(errstr,*) "Step: ", i, "out of ",xdimo
             call message(AppMsg,Sparse)
          endif
          do ii=1,ydimo
             if(expose(i,ii) .gt. 0.) then
                b   = 0.
                c   = 0.
                e   = 0.
                num = 0
                k   = -1
                rad = 0.0
                do while (c .lt. smoofact)
                   blast = b
                   clast = c
                   elast = e
                   numlast = num
                   k = k + 1
                   if(k .eq. 1) then
                      rad = 1.5*1.5
                   elseif(k .eq. 2) then
                      rad = 2.5*2.5
                   elseif(k .eq. 3) then
                      rad = 3.2*3.2
                   else
                      rad = k*k*1.05*1.05
                   endif
                   j1 = i - k
                   if(j1 .lt. 1) j1 = 1
                   j2 = i + k
                   if(j2 .gt. xdimo) j2 = xdimo
                   jj1 = ii - k
                   if(jj1 .lt. 1) jj1 = 1
                   jj2 = ii + k
                   if(jj2 .gt. xdimo) jj2 = xdimo
                   b = 0.
                   c = 0.
                   e = 0.
                   num = 0
                   do j=j1,j2
                      do jj=jj1,jj2
                         trad = (j-i)**2 + (jj-ii)**2
                         if((expose(j,jj) .gt. 0.) .and. &
                              (trad .le. rad)) then
                            if(rad .eq. 0.0) then
                               b = b + back(j,jj)
                               c = c + counts(j,jj)
                               e = e + expose(j,jj)
                            else
                               t = exp(-((trad/rad)**2))
                               b = b + back(j,jj) * t
                               c = c + counts(j,jj) * t
                               e = e + expose(j,jj) * t
                            endif
                            num = num + 1
                         endif
                      enddo
                   enddo
                enddo
                if((k .gt. 1) .and. (numlast .gt. 0.)) then
                   !                write(72,*)"1",i,ii,sca,clast,blast,elast
                   temp = sca * (clast - blast) / elast
                   adapted(i,ii) = temp
                elseif(num .gt. 0.) then
                   !                write(72,*)"2",i,ii,sca,c,b,e
                   temp = sca * (c - b) / e
                   adapted(i,ii) = temp
                else
                   !                write(72,*)"3",i,ii,sca,c,b,e
                   adapted(i,ii) = temp
                endif
                rad = 2.5*sqrt(rad)/60.0             
                sized(i,ii) = rad
                ir = 15*rad + 1
                yh(ir) = yh(ir) + 1
             endif
          enddo
       enddo
    end if
    ! SS:  Get an unused lun
    ! BHP: do the SIZE plot
    ymax = 0.0
    do i=1,100
       if(yh(i) .gt. ymax) ymax = yh(i)
    enddo
    if (withsmoo) then 
       warndie = .FALSE.
       call check_clobber(outSIZEQ_F,warndie)
       outlun = GetLun()
       open(unit=outlun,status='new',form='formatted',&
            file=trim(adjustl(outSIZEQ_F)),iostat=ios)
       if (ios .ne. 0) then
          call warning("badQDP",&
               "QDP output file could not be opened for write, no SIZE QDP output done")
       else
          ymax = 1.05*ymax
          write(outlun,*) 'VIEWPORT 0.15 0.15 0.95 0.85'
          write(outlun,*) 'FONT ROMAN'
          write(outlun,*) 'CS 1.8'
          write(outlun,*) 'LA X Conical Sampling FWHM (arcmin)'
          write(outlun,*) 'LA Y Number'
          write(outlun,*) 'LA F ', trim(adjustl(outRADQ_F))
          write(outlun,*) 'COL 1 ON 2'
          write(outlun,*) 'R X 0 5'
          write(outlun,*) 'R Y 0 ', ymax
          write(outlun,*) 'LW 2'
          write(outlun,*) 'LA F'
          write(outlun,*) 'LA T Smoothing FWHM Histogram'
          write(outlun,*) 'LINE STEP'
          do i=1,1000
             write(outlun,*) xh(i), yh(i)
          enddo
          close(outlun)
       end if
    end if
    !
    ! BHP: Open the new file size.fits and adapt.fits, write out
    !
    cdelt(1) = binfact * cdeltE(1) ! exactly the same adapt/binimage
    cdelt(2) = binfact * cdeltE(2) ! for all 4 images 
    crpix(1) = xdimo / 2 + 0.5
    crpix(2) = ydimo / 2 + 0.5
    units = 'c/s/deg2'
    bzero = 0.0
    bscale = 1.0
    area = 60.0*cdeltE(2)
    !
    if (withsmoo) then 
       if(fill .gt. 0) then ! fill only done with adapt_merge???!!!
          write(errstr,*) "Starting infill process"
          call message(appmsg,sparse)
          do ib=1,fill
             write(errstr,2100) ib
             call message(appmsg,sparse)
2100         format(1x,'Starting infill pass: ',i3)
             do i=2,xdimo-1
                if(mod(i,100) .eq. 0) then
                   write(errstr,*) "Step ",i," out of ",xdimo
                   call message(appmsg,always)
                endif
                do ii=2,ydimo-1
                   if(adapted(i,ii) .eq. 0.0) then
                      t = 0.0
                      tn = 0.0
                      do j=i-1,i+1
                         do jj=ii-1,ii+1
                            if(adapted(j,jj) .ne. 0.0) then
                               t = t + adapted(j,jj)
                               tn = tn + 1
                            endif
                         enddo
                      enddo
                      if(tn .ge. 3.0) then
                         counts(i,ii) = t/tn
                      else
                         counts(i,ii) = 0.0
                      endif
                   else
                      counts(i,ii) = adapted(i,ii)
                   endif
                enddo
             enddo
             do i=2,xdimo-1
                do ii=2,ydimo-1
                   adapted(i,ii) = counts(i,ii)
                enddo
             enddo
             write(errstr,2200) ib
             call message(appmsg,sparse)
2200         format(1x,'infill pass',i3,' complete')
          enddo
       endif
       write(errstr,*) "Adapted Counts   : ", sum(adapted)
       call message(AppMsg,Always)
       write(errstr,*) "Size Counts      : ", sum(sized)
       call message(AppMsg,Always)
    endif
    !
    status = 0
    warndie = .TRUE.
    type = 3 ! 3 = image
    name = "PRIMARY"
    if (withsmoo) then 
       call check_clobber(outADAPT_F,warndie) ! croak if SAS_CLOBBER is set
       ! 3 = CREATE
       call check_open_set(outADAPT_F,outADAPT_S,3,status,warndie) ! goflib/gofio_mod
       outADAPT_A = addArray(outADAPT_S, trim(name), REAL32, (/xdimo,ydimo/))
       outADAPT_RP => real32Array2Data(outADAPT_A)    
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_clobber(outSIZE_F,warndie)
       ! 3 = CREATE
       call check_open_set(outSIZE_F,outSIZE_S,3,status,warndie) ! goflib/gofio_mod
       outSIZE_A = addArray(outSIZE_S, trim(name), REAL32, (/xdimo,ydimo/))
       outSIZE_RP => real32Array2Data(outSIZE_A)
    else
       call check_clobber(outRATE_F,warndie) ! croak if SAS_CLOBBER is set
       ! 3 = CREATE
       call check_open_set(outRATE_F,outRATE_S,3,status,warndie) ! goflib/gofio_mod
       outRATE_A = addArray(outRATE_S, trim(name), REAL32, (/xdimo,ydimo/))
       outRATE_RP => real32Array2Data(outRATE_A)    
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_clobber(outSIGMA_F,warndie)
       ! 3 = CREATE
       call check_open_set(outSIGMA_F,outSIGMA_S,3,status,warndie) ! goflib/gofio_mod
       outSIGMA_A = addArray(outSIGMA_S, trim(name), REAL32, (/xdimo,ydimo/))
       outSIGMA_RP => real32Array2Data(outSIGMA_A)
    endif
    !
    ! add comments from originals ???!!!
    !
    if (withsmoo) then
       call updateKeywords(outADAPT_A,radecsysE,crpix,ctypeE,crvalE,cdelt)
       call updateKeywords(outSIZE_A,radecsysE,crpix,ctypeE,crvalE,cdelt)
    else
       call updateKeywords(outRATE_A,radecsysE,crpix,ctypeE,crvalE,cdelt)
       call updateKeywords(outSIGMA_A,radecsysE,crpix,ctypeE,crvalE,cdelt)       
    endif
    !
    ! fill the final array pointers opened above
    !
    do i=1,xdimo
       do j=1,ydimo
          if (withsmoo) then 
             outADAPT_RP(i-1,j-1) = adapted(i,j)
             outSIZE_RP(i-1,j-1)  = sized(i,j)
          else
             outRATE_RP(i-1,j-1)  = rate(i,j)
             outSIGMA_RP(i-1,j-1) = sigma(i,j)
          endif
       enddo
    enddo
    ! close and seal the output files
    if (withsmoo) then 
       call release(outADAPT_S)
       call release(outSIZE_S)
    else
       call release(outRATE_S)
       call release(outSIGMA_S)
    endif
    
    return

  end subroutine binadaptmerge
  !
  ! Subroutine section
  !
  subroutine makeFileNames(prefix,suffix,style,celow,cehigh,file)
    !
    character(len=*), intent(in) :: prefix, suffix, style, celow, cehigh
    character(len=256), intent(out) :: file
    character(len=3) :: pre, sky
    !
    !
    file = trim(adjustl(prefix)) // "-" // trim(adjustl(style)) //'-' // &
         trim(adjustl(celow)) // '-' // trim(adjustl(cehigh)) // trim(adjustl(suffix))
    
    return

  end subroutine makeFileNames
  
  !
  subroutine updateKeywords(out_A,radecsys,crpix,ctype,crval,cdelt)

    type(ArrayT)   :: out_A
    character(len=10) :: radecsys, ctype(2)
    real(kind=double) :: crval(2), crpix(2), cdelt(2)

    call setAttribute(out_A, 'CONTENT', "IMAGE")
    call setAttribute(out_A, 'TELESCOP', "XMM", "mission name")
    call setAttribute(out_A, 'INSTRUME', "EPIC","instrument name")
    call setAttribute(out_A, 'OBS_MODE', "POINTING","obs mode: POINTING, SLEW, OR SCAN")
    call setAttribute(out_A, 'RADECSYS', radecsys,"Equitorial system reference")
    call setAttribute(out_A, 'EQUINOX', 2000, "Equinox")
    !call setAttribute(out_A, 'BZERO', 0., "") how to write???!!!
    !call setAttribute(out_A, 'BSCALE', 1., "")
    !call setAttribute(out_A, 'BUNIT', "c/s/deg2", "Units of data")
    call setAttribute(out_A, 'CRPIX1', crpix(1), "Reference pixel")
    call setAttribute(out_A, 'CTYPE1', ctype(1), "Projection")
    call setAttribute(out_A, 'CRVAL1', crval(1), "Central Longitude")
    call setAttribute(out_A, 'CDELT1', cdelt(1), "WCS pixel size")
    call setAttribute(out_A, 'CUNIT1', "deg", "Units of coordinate")
    call setAttribute(out_A, 'CRPIX2', crpix(2), "Reference pixel")
    call setAttribute(out_A, 'CTYPE2', ctype(2), "Projection")
    call setAttribute(out_A, 'CRVAL2', crval(2), "Central Latitude")    
    call setAttribute(out_A, 'CDELT2', cdelt(2), "WCS pixel size")
    call setAttribute(out_A, 'CUNIT2', "deg", "Units of coordinate")

    return

  end subroutine updateKeywords


end module binadaptmerge_mod
